"""Functions used to read in GMNS files and networks.

Typical Usage:

    ```python
    read_gmns_csv('csv_file',validate=True)
    read_gmns_network('gmns_dir_path')
    ```
"""

from os.path import dirname, join, realpath

import pandas as pd

from gmnspy.utils import logger
from gmnspy.validation import (
    apply_schema_to_df,
    check_required_files,
    update_resources_based_on_existence,
    validate_foreign_keys,
)

from .schema import read_config

spec_folder = join(dirname(realpath(__file__)), "spec")


def read_gmns_csv(filename: str, validate: bool = True, schema_file: str = None) -> pd.DataFrame:
    """
    Read csv and returns it as a dataframe.

    Optionally coerced to the types as specified in the data schema.

    Args:
        filename: file location of the csv file to read in.
        validate: boolean whether to apply the specified schema to the dataframe. Default is True.
        schema_file: file location of the schema to validate the file to.

    Returns: Validated dataframe with coerced types according to schema.
    """
    df = pd.read_csv(filename)

    if validate:
        apply_schema_to_df(df, schema_file=schema_file, originating_file=filename)
    else:
        logger.info(f"not validating {filename}")

    return df


def read_gmns_network(data_directory: str, config: str = None, raise_error=False) -> dict:
    """
    Read and validate each GMNS file as specified in the config.

    Validation includes foreign keys between the tables.

    Args:
        data_directory: Directory where GMNS data is.
        config: Configuration file. A json file with a list of "resources"
            specifying the "name", "path", and "schema" for each GMNS table as
            well as a boolean value for "required". If not specified, assumes
            it is in a subdirectory "spec/gmns.spec.json"
        raise_error: Raises error if missing folder

            Example:
            ::
                {
                  "resources": [
                   {
                     "name":"link",
                     "path": "link.csv",
                     "schema": "link.schema.json",
                     "required": true
                   },
                   {
                     "name":"node",
                     "path": "node.csv",
                     "schema": "node.schema.json",
                     "required": true
                   }
                 }
    returns: a dictionary mapping the name of each GMNS table to a
        validated dataframe.
    """
    config = config or join(spec_folder, "gmns.spec.json")
    gmns_net_d = {}

    resource_df = read_config(config, data_dir=data_directory) if config else 1

    # check required files exist,
    check_required_files(resource_df, raise_error)

    # update resource dictionary based on what files are in the directory
    resource_df = update_resources_based_on_existence(resource_df)

    # read each csv to a df and validate format
    # todo add paired schema
    for _, row in resource_df.iterrows():
        gmns_net_d[row["name"]] = read_gmns_csv(row["fullpath"])

    # validate foreign keys
    validate_foreign_keys(gmns_net_d, resource_df, raise_error)

    return gmns_net_d



"""
Functions related to Frictionless Data Schemas for GMNS.

Typical usage:

    ```python
    read_schema(schema_file)
    read_config(config_file)
    document_schemas_to_md(schema_file_dir)
    document_spec_to_md(spec_file)
    ```
"""

import glob
import json
from os.path import dirname, join, realpath
from pathlib import Path

import frictionless
import pandas as pd

from .utils import list_to_md_table, logger

SCHEMA_TO_PANDAS_TYPES = {
    "integer": "Int64",
    "number": "Float64",
    "string": "string",
    "any": "object",
    "boolean": "boolean",
}

FORMAT_TO_REGEX = {
    # https://emailregex.com/
    "email": r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$",
    # https://www.regextester.com/94092
    "uri": r"^\w+:(\/?\/?)[^\s]+$",
}


def read_schema_for_resource(resource_df: pd.DataFrame, table_name: str, raise_error: bool) -> dict:
    """
    Read in schema from schema json file and returns as dictionary.

    ##TODO validate schema itself

    Args:
        schema_file: File location of the schema json file.

    Returns: The schema as a dictionary
    """
    matching_schema_paths = resource_df.loc[resource_df["name"] == table_name, "fullpath_schema"]
    if matching_schema_paths.empty:
        msg = f"FAIL. Could not find schema path for table {table_name}"
        logger.error(msg)
        if raise_error:
            raise Exception(msg)
        return dict()
    return read_schema(schema_file=matching_schema_paths.iloc[0])


def read_schema(schema_file: str) -> dict:
    """
    Read in schema from schema json file and returns as dictionary.

    ##TODO validate schema itself

    Args:
        schema_file: File location of the schema json file.

    Returns: The schema as a dictionary
    """
    with open(schema_file, encoding="utf-8") as f:
        schema = json.load(f)
    return schema


def read_config(config_file: str, data_dir: str = "", schema_dir: str = "") -> pd.DataFrame:
    """
    Read a GMNS config file, adds some full paths and returns as a dataframe.

    Args:
        config_file: Configuration file. A json file with a list of "resources"
            specifying the "name", "path", and "schema" for each GMNS table as
            well as a boolean value for "required".
            Example:
            ::
                {
                  "resources": [
                   {
                     "name":"link",
                     "path": "link.csv",
                     "schema": "link.schema.json",
                     "required": true
                   },
                   {
                     "name":"node",
                     "path": "node.csv",
                     "schema": "node.schema.json",
                     "required": true
                   }
                 }
        data_dir: Directory where GMNS files are. If not specified, assumes
            the same directory as the config_file.
        schema_dir: Directory where GMNS schema files are. If not specified, assumes
            the same directory as the config_file.

    Returns: GMNS configuration file as a DataFrame.
    """
    with open(config_file, "r", encoding="utf-8") as f:
        config = json.load(f)
    ## todo validate config

    resource_df = pd.DataFrame(config["resources"])
    resource_df["required"] = resource_df["required"].fillna(False)

    logger.info(str(resource_df))

    # Add full paths to data files
    if not data_dir:
        data_dir = dirname(config_file)
    resource_df["fullpath"] = resource_df["path"].apply(lambda x: join(data_dir, x))

    # Add full paths to data files
    if not schema_dir:
        schema_dir = dirname(config_file)
    resource_df["fullpath_schema"] = resource_df["schema"].apply(lambda x: join(schema_dir, x))
    logger.info(str(resource_df))

    resource_df = resource_df.set_index("name", drop=False)
    return resource_df


def document_schemas_to_md(schema_path: str = None, out_path: str = None) -> str:
    """Create markdown for each **.schema.json file in schema_path.

    Args:
        schema_path (str, optional): Path to look for schema files.
            Defaults to join(dirname(realpath(__file__)), "spec")
        out_path (str, optional): If specified, will write out resulting markdown to this file.
            Defaults to None.

    Returns:
        str: Markdown string
    """
    schema_path = schema_path or join(dirname(realpath(__file__)), "spec")
    logger.info(f"Documenting Schemas in:\n {schema_path}")

    schema_files = glob.glob(join(schema_path, "**/*.schema.json"), recursive=True)

    # Create markdown with a table for each schema file
    schema_markdown = ""

    for sf in schema_files:
        logger.info(f"Adding to MD: {sf}")
        s = frictionless.Schema(sf)
        md = s.to_markdown()
        _name = f"## {Path(sf).stem.split('.')[-2]}"
        md = md.replace("## `schema`", _name)

        schema_markdown += f"\n{md}\n"

    if out_path:
        with open(out_path, "w") as f:
            f.write(str(schema_markdown))

    return schema_markdown


def document_spec_to_md(spec_path: str = None, out_path: str = None) -> str:
    """Create markdown for each **.schema.json file in schema_path.

    Args:
        spec_path (str, optional): Path to look for spec file.
            Defaults to join(dirname(realpath(__file__)), "**/gmns.spec.json")
        out_path (str, optional): If specified, will write out resulting markdown to this file.
            Defaults to None.

    Returns:
        str: Markdown string
    """
    DROP_COLS = ["fullpath", "fullpath_schema", "path", "schema", "name"]

    spec_path = spec_path or join(dirname(realpath(__file__)), "spec", "gmns.spec.json")

    logger.info(f"Documenting Spec in:\n {spec_path}")

    # Generate a table for overall file requirements
    spec_df = read_config(spec_path)
    spec_df = spec_df.drop(columns=DROP_COLS).reset_index()
    spec_df["name"] = spec_df["name"].apply(lambda x: f"[`{x}`](#{x})".replace("_", "-"))

    spec_markdown = spec_df.to_markdown(index=False)

    if out_path:
        with open(out_path, "w") as f:
            f.write(str(spec_markdown))

    return spec_markdown



"""
Functions which test GMNS schema constraints.

Represented by functions with naming pattern `_<constraint_name>_constraint`
    which take a series and a single parameter as input.

Constraints are specified in the gmns spec files for each field are treated
    as mandatory. The same parameters can be specified as warnings and
    will not be treated as mandatory.
"""

from typing import Union

import pandas as pd


def _required_constraint(_s, _p) -> Union[None, str]:
    """Placeholder for something tested somewhere else."""
    pass


def _unique_constraint(s: pd.Series, _) -> Union[None, str]:
    """
    Check if series contains unique values.

    TODO: ##needstest

    Args:
        s: series that shouldn't exceed maximum value.
        _: boolean specifying unique values needed.

    Returns:
        An error string if there is an error. Otherwise, None.
    """
    if s.dropna().duplicated():
        return "Values not unique."


def _minimum_constraint(s: pd.Series, minimum: Union[float, int]) -> Union[None, str]:
    """
    Check if series contains value under the specified minimum.

    TODO: ##needstest

    Args:
        s: series that shouldn't be under the minimum value.
        minimum: minimum value for the series.

        Returns:
        An error string if there is an error. Otherwise, None.
    """
    if s[s < minimum].dropna().to_list():
        return "Values lower than minimum: {}".format(minimum)


def _maximum_constraint(s: pd.Series, maximum: Union[float, int]) -> Union[None, str]:
    """
    Check if series contains value above the specified maximum.

    TODO: ##needstest

    Args:
        s: series that shouldn't exceed maximum value.
        maximum: maximum value for the series.

    Returns:
        An error string if there is an error. Otherwise, None.
    """
    if s[s > maximum].dropna().to_list():
        return "Values higher than maximum: {}".format(maximum)


def _pattern_constraint(s: pd.Series, pattern: str) -> Union[None, str]:
    """
    Check if series contains values conforming to specified pattern.

    TODO: ##needstest

    Args:
        s: series that shouldn't be under the minimum value.
        pattern: regex string.

    Returns:
        An error string if there is an error. Otherwise, None.
    """
    if ~s.str.contains(pattern):
        return "Doesn't match pattern: {}".format(pattern)


def _enum_constraint(s: pd.Series, enum: Union[str, list], sep: str = ",") -> Union[None, str]:
    """
    Check if series contains valid enum values.

    TODO: ##needstest

    Args:
        s: series of values that should all have values in  the enumerated
            list
        enum: either a string of allowable values separated by sep, or
            a list of allowable values.
        sep: separator for different values. Default is ","

    Returns:
        An error string if there is an error. Otherwise, None.
    """
    if not isinstance(enum, list):
        enum = enum.split(sep)
    err_i = (s[~s.isin(enum)]).to_list()
    if err_i:
        return "Values: {} not in enumerated list: {}".format(err_i, enum)



"""GMNSpy package.

Typical usage:

    ```python
    import gmnspy
    ```
"""

from .in_out import read_gmns_csv, read_gmns_network
from .schema import (
    document_schemas_to_md,
    document_spec_to_md,
    read_config,
    read_schema,
)
from .utils import list_to_md_table, logger

__all__ = [
    "read_gmns_csv",
    "read_gmns_network",
    "document_schemas_to_md",
    "document_spec_to_md",
    "read_config," "read_schema",
    "list_to_md_table",
    "logger",
]



"""
Establishes logging parameters.

Logger is run at import.
"""
import logging
from os.path import join
from tempfile import gettempdir


def build_logger():
    """Build logger object.

    Returns: Logger object.
    """
    logger = logging.getLogger("gmnspy")
    handler_name = "gmnspy"

    if len([h for h in logger.handlers if h.name and handler_name == h.name]) > 0:
        return logger

    log_path = join(gettempdir(), "gmnspy.log")
    FORMATTER = logging.Formatter("%(asctime)s;%(levelname)s ; %(message)s", datefmt="%H:%M:%S:")
    ch = logging.FileHandler(log_path)
    ch.setFormatter(FORMATTER)
    ch.set_name(handler_name)
    ch.setLevel(logging.INFO)
    logger.addHandler(ch)

    return logger


logger = build_logger()



"""Function which establishes which GMNS resources are present."""

from os.path import exists

import pandas as pd

from gmnspy.utils.set_logger import logger


def update_resources_based_on_existence(resource_df: pd.DataFrame) -> pd.DataFrame:
    """
    Update resource dataframe based on which files exist in the directory.

    Args:
        resource_df:Dataframe with a row for each GMNS table including
            the the columns "fullpath" and "name".

    Returns: Updated version of resource dataframe without non-existent
        files.
    """
    updated_resource_df = resource_df[resource_df["fullpath"].apply(lambda x: exists(x))]

    logger.info(f"""Found following files to define network: \n - {updated_resource_df["name"].to_list()}""")
    return updated_resource_df



"""Applies schema constraints to update a dataframe of data."""

import os
from os.path import dirname, join, realpath

import pandas as pd

from gmnspy.schema import SCHEMA_TO_PANDAS_TYPES, read_schema
from gmnspy.utils import logger
from gmnspy.validation.constraint_checking import (
    _enum_constraint,
    _maximum_constraint,
    _minimum_constraint,
    _pattern_constraint,
    _required_constraint,
    _unique_constraint,
)


def apply_schema_to_df(
    df: pd.DataFrame, schema_file: str = None, originating_file: str = None, raise_error=False
) -> pd.DataFrame:
    """
    Evaluate a gmns table against a specified data schema.

    (1) Checks required fields exist
    (2) Coerces field types
    (3) Evaluates constraints that are firmly enforced
    (4) Evaluates warnings for strange values

    Args:
        df: Dataframe of the gmns table
        schema_file: The table schema that will be applied.  If not supplied,
            will look for the schema in the directory that matches the name in
            originating_file (i.e. link, node)

        originating_file: base name of the originating file (i.e. link.csv, node)
        raise_error: Raises error if error found

    Returns:
        DataFrame with fields coerced to appropriate type and constraints
            and warnings evaluated.
    """
    if not schema_file:
        schema_filename = os.path.split(originating_file)[-1].split(".")[0] + ".schema.json"
        schema_file = join(join(dirname(realpath(__file__)), "../spec"), schema_filename)
    logger.info("SCHEMA: {}".format(schema_file))
    logger.info("...validating {} against {}".format(df, schema_file))
    schema = read_schema(schema_file=schema_file)

    """
    1. Check field names and requirements
    - Required fields present (strict)
    - Extra fields that aren't in spec (warn)
    """
    logger.debug(schema)

    required_fields = [f["name"] for f in schema["fields"] if f.get("constraints", {}).get("required")]
    missing_required_fields = set(required_fields) - set(df.columns)
    if missing_required_fields:
        msg = f"FAIL. Missing required fields {missing_required_fields}"
        logger.error(msg)
        if raise_error:
            raise Exception(msg)

    fields = [f["name"] for f in schema["fields"]]
    extra_fields = set(df.columns) - set(fields)
    if extra_fields:
        logger.warning(f"WARN. Extra fields outside of spec: {extra_fields}")

    """
    2. Coerce types

    ##TODO: enforce formats
    """
    used_fields = [f["name"] for f in schema["fields"] if f["name"] in df.columns]
    types = [SCHEMA_TO_PANDAS_TYPES[f["type"]] for f in schema["fields"] if f["name"] in df.columns]
    # fmt = [FORMAT_TO_REGEX.get(f.get("format", None), None) for f in schema["fields"]]

    field_types = dict(zip(used_fields, types))
    # field_formats = dict(zip(used_fields, fmt))

    # print(field_types)
    try:
        df = df.astype(field_types)
        logger.info("Passed field type coercion")
    except Exception as e:
        logger.critical(f"ouch. {e.args}")
        if raise_error:
            raise Exception(f"ouch. {e.args}")

    """
    3. Check field constraints
    """
    fields_with_constraints = [f["name"] for f in schema["fields"] if f["name"] in df.columns and f.get("constraints")]
    constraints = [f["constraints"] for f in schema["fields"] if f["name"] in df.columns and f.get("constraints")]

    # iterate through all the constraints for all the fields
    error_list = []

    for field_name, fld_constr in zip(fields_with_constraints, constraints):
        error_list += [globals()[f"_{c_name}_constraint"](df[field_name], cpar) for c_name, cpar in fld_constr.items()]
    error_list = [i for i in error_list if i]

    if error_list:
        logger.error(error_list)
        if raise_error:
            raise Exception(error_list)
    else:
        logger.info("Passed Field Required Constraint Validation")

    """
    4. Check field warnings
    """
    fields_with_warnings = [f["name"] for f in schema["fields"] if f["name"] in df.columns and f.get("warnings")]
    warnings = [f["warnings"] for f in schema["fields"] if f["name"] in df.columns and f.get("warnings")]

    warning_list = []
    for field_name, field_warnings in zip(fields_with_warnings, warnings):
        # print(field_name,field_constraints)
        warning_list += [
            globals()["_" + c_name + "_constraint"](df[field_name], c_param)
            for c_name, c_param in field_warnings.items()
        ]
    warning_list = [i for i in error_list if i]

    logger.info(warning_list if warning_list else "No Field Warnings")
    return df



"""Functions which test GMNS foreign key constraints."""

from typing import Dict

import pandas as pd

from gmnspy.schema import read_schema_for_resource
from gmnspy.utils import logger


def validate_foreign_keys(gmns_net_d: Dict[str, pd.DataFrame], resource_df: pd.DataFrame, raise_error) -> None:
    """
    Validate foreign keys for each GMNS table.

    Finds foreign keys in schemas of each GMNS table and validates that
    they link to a valid series which has (1) unique IDs, and (2) contains
    the values of the referring series.

    Args:
        gmns_net_d: Dictionary containing dataframes of all the GMNS tables
            for the network keyed to their file names (i.e. "link").
        resource_df:Dataframe with a row for each GMNS table which contains
            the field "fullpath_schema" for the schema locations for
            each GMNS table which is where foreign keys are specified.
        raise_error: Raises error if found
    """
    logger.info(gmns_net_d["node"]["node_id"])

    fkey_errors = []
    for table_name, df in gmns_net_d.items():
        schema = read_schema_for_resource(resource_df, table_name, raise_error)
        if len(schema) == 0:
            continue

        foreign_keys = [
            (f["name"], f["foreign_key"])
            for f in schema["fields"]
            if (f.get("foreign_key") and f["name"] in df.columns)
        ]
        logger.debug("FKEYS: ", foreign_keys)

        # find the series for the foreign keyl
        for field, f_key in foreign_keys:
            # NOTE this requires that field names that are foreign keys not have '.'
            t, f = f_key.split(".")
            if not t:
                reference_s = df[f]
            else:
                reference_s = gmns_net_d.get(t, {}).get(f, None)
                if reference_s is None:
                    msg = f"FAIL. {f} field in table {t} does not exist"
                    logger.error(msg)
                    if raise_error:
                        raise Exception(msg)
                    continue
            fkey_errors += validate_foreign_key(df[field], reference_s, raise_error)


def validate_foreign_key(source_s: pd.Series, reference_s: pd.Series, raise_error: bool) -> list:
    """
    Validate foreign keys of single pair of source and reference series.

    Checks that the source_s series links to a valid reference_s series
        which has (1) unique IDs, and (2) contains the values of the
        referring series.

    source_s: series containing values that reference foreign key values in reference_s.
    reference_s: series of foreign key values.
    raise_error: Raises error if found

    Returns: a list of error messages.
    """
    fkey_errors = []
    # Make sure reference_s is unique
    dupes = reference_s.dropna().duplicated()
    if dupes.any():
        msg = "FAIL. Duplicates exist in foreign key series: {}".format(dupes)
        logger.error(msg)
        if raise_error:
            raise Exception(msg)
        fkey_errors.append(msg)

    # Make sure all source have a valid reference
    if not source_s.isin(reference_s.dropna().to_list()).any():
        msg = "FAIL. {} not in foreign key reference.".format(source_s[source_s.isin(reference_s.dropna().to_list())])
        logger.error(msg)
        if raise_error:
            raise Exception(msg)
        fkey_errors.append(msg)
    return fkey_errors



"""Utility functions for formatting tables in markdown."""


def list_to_md_table(list_of_dicts: list) -> str:
    """
    Read a list of dictionaries defining a schema and creates a markdown table.

    NOTE: currently unused.

    args:
        list_of_dicts: a list of dictionaries containing definition of fields.

    returns: A markdown string representing the table.

    """
    # flatten dictionary
    flat_list = []
    for i in list_of_dicts:
        flat_list.append({k: v for k, v in _recursive_items(i)})

    # get all items listed
    standard_header_items = [
        "name",
        "required",
        "type",
        "foreign_key",
        "description",
    ]
    additional_header_items = list(set([k for i in flat_list for k, v in i.items()]) - set(standard_header_items))
    header_items = standard_header_items + sorted(additional_header_items)

    header_md = "|" + "|".join(header_items) + "|\n"
    header_md += "|---" * len(header_items) + "|\n"

    body_md = ""
    for d in flat_list:
        body_md += "|" + "|".join([_format_cell_by_type(d.get(i, "-"), i) for i in header_items]) + "|\n"

    return header_md + body_md


def _format_cell_by_type(x, header):
    if "enum" in header and len(x) > 1:
        return "Allowed Values: `" + ",".join(map(str, x)) + "`"
    if header == "foreign_key" and len(x) > 1:
        table, var = x.split(".")
        if table == "":
            return "Variable: `{}`".format(var)
        return "Table: `{}`, Variable: `{}`".format(table, var)
    else:
        return str(x)


def _recursive_items(d: dict, key_prefix: str = "") -> dict:
    """
    Recursively flattens a nested dictionary to a dictionary.

    Modified from https://stackoverflow.com/questions/39233973/get-all-keys-of-a-nested-dictionary

    Returns: a dictionary with key_prefixe to represent original nesting structures.
    """
    for k, v in d.items():
        if type(v) is dict:
            yield from _recursive_items(v, key_prefix=k + " ")
        elif k == "required":
            yield (k, v)
        else:
            yield (key_prefix + k, v)



"""Functions which test GMNS file requirement constraints."""

from os.path import exists

import pandas as pd

from gmnspy.utils import logger


def check_required_files(resource_df: pd.DataFrame, raise_error=False) -> None:
    """
    Check required files exist. Will fail if they don't.

    Args:
        resource_df: Dataframe with a row for each GMNS table including
            the the columns "fullpath", "name", and "required" (boolean).
        raise_error: Raises error if missing folder
    """
    req_files = resource_df.query("required == True")

    missing_required_files = req_files[req_files["fullpath"].apply(lambda x: not exists(x))][["name", "fullpath"]]

    if not missing_required_files.shape[0]:
        return

    if raise_error:
        msg = f"FAIL Missing Required Files: {missing_required_files}"
        logger.error(msg)
        raise Exception(msg)



"""Submodule with validation checks."""

from .foreign_keys import validate_foreign_keys
from .required_files import check_required_files
from .resources_existance import update_resources_based_on_existence
from .schema_to_df import apply_schema_to_df



import glob
from os.path import dirname, join, realpath

import pytest

import gmnspy

base_path = dirname(dirname(realpath(__file__)))

# https://github.com/zephyr-data-specs/GMNS/edit/master/Small_Network_Examples/Multiple_Bike_Facilities/road_link.csv

test_data = [
    "link",
    "geometry",
    "node",
    "use_definition",
    "use_group",
]
schemas = glob.glob("../gmnspy/**/*.schema.json", recursive=True)
test_pth = join(dirname(realpath(__file__)), "data")


def test_read_schema():
    schema_file = join(base_path, "gmnspy", "spec", "link.schema.json")
    _ = gmnspy.read_schema(schema_file)


@pytest.mark.parametrize("test_data_name", test_data)
def test_validate_dfs(test_data_name):
    _ = gmnspy.in_out.read_gmns_csv(join(test_pth, f"{test_data_name}.csv"))


def test_validate_relationships():
    _ = gmnspy.in_out.read_gmns_network(join(base_path, "tests", "data"))


@pytest.mark.parametrize("schema_file", schemas)
def test_read_schema2(schema_file):
    # schema_file = join(base_path, "spec", "link.schema.json")
    s = gmnspy.read_schema(schema_file)
    _ = gmnspy.list_to_md_table(s["fields"])
